"""Parameter reconstruction using JCMsuite

The details of the employed reconstruction are described in [1].

[1] https://onlinelibrary.wiley.com/doi/10.1002/adts.202200112

"""

import pathlib
import pandas as pd

import jcmwave
import numpy as np
from matplotlib import pyplot as plt
from utils.forward_problem import ForwardProblem

# from utils.generate_test_data import generate_test_data
from utils.material import SiMaterial

##############################################################################
# Parameters to play with

# Number of parallel jobs in JCMsolve; for demo version only 1 is allowed
MULTIPLICITY: int = 1

# Using derivatives of the FEM model for the reconstruction; speeds up the reconstruction
DERIVATIVE_ORDER: int = 1

# FEM degree of the JCMsuite model
FEM_DEGREE: int = 3

# Number of iterations in the optimization
NUM_ITERATIONS: int = 20

##############################################################################

# Register the localhost as a machine to perform computations
jcmwave.daemon.add_workstation(
    Hostname="localhost",
    #
    # Number of parallel jobs
    Multiplicity=MULTIPLICITY,
    #
    # Each parallel job runs with this number of threads. Use at least 2.
    NThreads=2,
)


def main():

    # Path to JCM project
    root_dir = pathlib.Path(__file__).parent
    data_dir = root_dir / "data"
    jcm_project_dir = root_dir / "jcm"

    project_file = jcm_project_dir / "project.jcmpt"

    optimization_dir = pathlib.Path(__file__).parent
    optimization_dir_str = str(optimization_dir.absolute())

    study_id = "Ellipsometry_reconstruction_example"

    # Remove any previous results
    (optimization_dir / f"{study_id}.jcmo").unlink(missing_ok=True)

    target_keys = pd.read_csv(data_dir / "target_parameters.csv")
    target_keys = pd.Series(
        target_keys.Value.values, index=target_keys.Parameter
    ).to_dict()

    target_values = pd.read_csv(data_dir / "target_values.csv")
    target_vector = target_values["target_vector"].to_numpy()
    uncertainty_vector = target_values["uncertainty_vector"].to_numpy()

    # Model parameters
    model_param_keys = dict(
        # Varying geometry
        h=55,
        width=31,
        swa=88,
        radius=8,
    )

    # Rest of the parameters for the JCMsuite project
    keys = dict(
        # Numerical parameters
        derivative_order=DERIVATIVE_ORDER,
        fem_degree=FEM_DEGREE,
        precision=1e-3,
        # Material parameters; n3 is overridden by the material file
        n1=1,
        n2=1.4,
        n3=1.967 + 4.443 * 1j,
        # Illumination
        theta=65,
        phi=45,
        vacuum_wavelength=365e-9,
    )

    # Merge the two
    keys.update(model_param_keys)

    wavelengths = np.linspace(266, 800, 11)

    fem_problem = ForwardProblem(project_file, wavelengths, SiMaterial())

    # Four dimensions
    optimization_domain = [
        {"name": "h", "type": "continuous", "domain": [50, 60]},
        {"name": "width", "type": "continuous", "domain": [25, 35]},
        {"name": "swa", "type": "continuous", "domain": [84, 90]},
        {"name": "radius", "type": "continuous", "domain": [6, 8]},
    ]

    print()
    print("The target parameter to be reconstructed is")
    for parameter in optimization_domain:
        print(f"\t{parameter['name']}: {target_keys[parameter['name']]}")
    print()
    print(
        "Derivative information of the FEM solver is {}".format(
            "being used" if keys["derivative_order"] > 0 else "not being used"
        )
    )
    print()

    client = jcmwave.optimizer.client()

    # Creation of the study object with study_id 'BayesLeastSquare_example'
    study = client.create_study(
        domain=optimization_domain,
        driver="BayesLeastSquare",
        name="Ellipsometry reconstruction example",
        study_id=study_id,
        save_dir=optimization_dir_str,
        open_browser=True,
    )

    # Definition of the objective function including derivatives
    def objective(**kwargs):

        objective_keys = keys.copy()
        objective_keys.update(kwargs)

        mueller_matrix, mueller_matrix_derivatives = fem_problem.solve(objective_keys)

        observation = study.new_observation()

        flat_mueller_matrix = mueller_matrix.flatten()
        observation.add(flat_mueller_matrix)

        if objective_keys["derivative_order"] > 0:
            for parameter in optimization_domain:
                if parameter["type"] == "continuous":
                    p = parameter["name"]
                    derivative_value = mueller_matrix_derivatives[p].flatten()
                    observation.add(derivative=p, value=derivative_value)

        return observation

    # Set study parameters
    study.set_parameters(
        target_vector=target_vector.tolist(),
        uncertainty_vector=uncertainty_vector.tolist(),
        max_iter=NUM_ITERATIONS,
    )

    # Run the minimization
    study.set_objective(objective)
    study.run()

    # Plot the reconstruction results and compare it to the target
    # First reshape the target vector into a 4x4 matrix
    target_matrix = target_vector.reshape(len(wavelengths), 4, 4)
    uncertainty_matrix = uncertainty_vector.reshape(len(wavelengths), 4, 4)

    # Update the keys with the minimum parameters and get the measurement data
    study_info = study.info()
    min_params = study_info["min_params"]
    keys.update(min_params)

    print("Generating reconstruction data for comparison")
    reconstructed_mueller_matrix, _ = fem_problem.solve(keys)

    fig, ax = plt.subplots(4, 4, figsize=(10, 10), sharex=True)

    fig.suptitle("Mueller matrix entries")

    for i in range(4):
        for j in range(4):

            # Set labels
            if i == 0 and j == 0:
                target_label = dict(label="Target")
                reconstructed_label = dict(label="Reconstructed")
            else:
                target_label = dict()
                reconstructed_label = dict()

            # Plot data
            ax[i, j].plot(wavelengths, target_matrix[:, i, j], **target_label)
            ax[i, j].plot(
                wavelengths,
                reconstructed_mueller_matrix[:, i, j],
                **reconstructed_label,
            )

            ax[i, j].set_title(f"M{i + 1}{j + 1}")
            if i == 3:
                ax[i, j].set_xlabel("Wavelength (nm)")

    plt.figlegend(loc="center", bbox_to_anchor=(0.77, 0.97))
    plt.tight_layout()
    plt.savefig(optimization_dir / "reconstruction.pdf", bbox_inches="tight")


if __name__ == "__main__":
    main()
